(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_slave.ml 239 2010-06-23 16:49:03Z gerd $ *)

(** The main module for the slaves. Defines RPC services
    for [Nameslave], and [Coordination].

    The programs [Filesystem] and [Dn_admin] are also "implemented", but all
    calls  (except [null]) fail.
 *)

(** Implementation of [Nameslave] *)

module Nn_db_transaction_var =
  Netplex_cenv.Make_var_type(struct type t = Nn_db.transaction option end)

let get_trans() =
  try
    Nn_db_transaction_var.get "current_transaction"
  with
    | _ -> None


let set_trans t =
  Nn_db_transaction_var.set "current_transaction" (Some t)


let unset_trans () =
  Nn_db_transaction_var.set "current_transaction" None


let proc_begin_transaction conf sess (cname,exp_rev) emit =
  if cname <> conf#nn_clustername then
    failwith "Nameslave.begin_transaction: bad clustername";
  if get_trans() <> None then
    failwith "Nameslave.begin_transaction: Already in a transaction";
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let t = Nn_db.transact esys in
  let e = Nn_db.revision_get_e t#connection esys in
  Uq_engines.when_state
    ~is_done:(fun rev ->
		if rev = exp_rev then (
		  set_trans t;
		  emit ()
		)
		else (
		  Netlog.logf `Err
		    "Revision string does not have the expected value";
		  Netlog.logf `Err
		    "Refusing to start transaction";
		  t#return();
		  Rpc_server.reply_error sess Rpc.System_err
		)
	     )
    ~is_error:(fun err ->
		 Netlog.logf `Err
		   "begin_transaction: Error %s"
		   (Netexn.to_string err);
		 t#return();
		 Rpc_server.reply_error sess Rpc.System_err
	      )
    e


let proc_prepare_commit sess () emit =
  match get_trans() with 
    | None ->
	failwith "Nameslave.prepare_commit: Not in a transaction"
    | Some t ->
	let e =
	  new Uq_engines.map_engine 
	    ~map_done:(fun () -> `Done true)
	    ~map_error:(fun error ->
			  Netlog.logf `Err
			    "Nameslave.prepare_commit: Error %s"
			    (Netexn.to_string error);
			  `Done false)
	    ~map_aborted:(fun () ->
			    Netlog.logf `Err
			      "Nameslave.prepare_commit: Aborted";
			    `Done false)
	    (t#prepare()) in
	Uq_engines.when_state
	  ~is_done:(fun ok ->
		      try emit ok
		      with Rpc_server.Connection_lost -> ()
		   )
	  e
	

let proc_commit sess () emit =
  match get_trans() with 
    | None ->
	failwith "Nameslave.commit: Not in a transaction"
    | Some t ->
	let e = t#commit() in
	Uq_engines.when_state
	  ~is_done:(fun () ->
		      unset_trans();
		      try emit()
		      with Rpc_server.Connection_lost -> ()
		   )
	  ~is_error:(fun error ->
		       Netlog.logf `Err
			 "Nameslave.commit: Error %s"
			 (Netexn.to_string error);
		       Rpc_server.reply_error sess Rpc.System_err)
	  ~is_aborted:(fun () ->
			 Netlog.logf `Err
			   "Nameslave.commit: Aborted";
			 Rpc_server.reply_error sess Rpc.System_err)
	  e


let push procname f sess emit =
  match get_trans() with 
    | None ->
	failwith ("Nameslave." ^ procname ^ ": Not in a transaction")
    | Some t ->
	let esys = (Netplex_cenv.self_cont()) # event_system in
	let e =
	  f t#connection esys in
	Uq_engines.when_state
	  ~is_done:(fun () ->
		      try emit ()
		      with Rpc_server.Connection_lost -> ()
		   )
	  ~is_error:(fun error ->
		       Netlog.logf `Err
			 "Nameslave.%s: Error %s"
			 procname (Netexn.to_string error);
		       Rpc_server.reply_error sess Rpc.System_err)
	  ~is_aborted:(fun () ->
			 Netlog.logf `Err
			   "Nameslave.%s: Aborted" procname;
			 Rpc_server.reply_error sess Rpc.System_err)
	  e


(* CHECK: maybe we can optimize here a bit - by sending the void result
   immediately. We would then synchronize first at prepare_commit time.
 *)

let proc_push_datastore_upd sess (id,identity,size,enabled) emit =
  push
    "push_datastore_upd"
    (Nn_db.datastore_upd_e ~id ~identity ~size ~enabled)
    sess
    emit
    

let proc_push_datastore_del sess id emit =
  push
    "push_datastore_del"
    (Nn_db.datastore_del_e ~id)
    sess
    emit


let proc_push_revision_upd sess revstr emit =
  push
    "push_revision_upd"
    (Nn_db.revision_upd_e ~revstr)
    sess
    emit

let proc_push_blockalloc_upd sess (ds_id,blkidx,blkmap) emit =
  push
    "push_blockalloc_upd"
    (Nn_db.blockalloc_upd_e ~datastore:ds_id ~blkidx ~blkmap)
    sess
    emit

let proc_push_inode_ins sess (id,ii) emit =
  push
    "push_inode_ins"
    (Nn_db.inode_ins_e ~id ii)
    sess
    emit

let proc_push_inode_upd sess (id,ii) emit =
  push
    "push_inode_upd"
    (Nn_db.inode_upd_e ~id ii)
    sess
    emit

let proc_push_inode_upd_time sess (id,mtime,ctime) emit =
  push
    "push_inode_upd_time"
    (Nn_db.inode_upd_time_e ~id ~mtime ~ctime)
    sess
    emit

let proc_push_inode_del sess id emit =
  push
    "push_inode_del"
    (Nn_db.inode_del_e ~id)
    sess
    emit

let proc_push_inodeblocks_ins sess (inode,bi_list) emit =
  push
    "push_inodeblocks_ins"
    (Nn_db.inodeblocks_ins_e ~inode (Array.to_list bi_list))
    sess
    emit

let proc_push_inodeblocks_del sess (inode,blkidx,len) emit =
  push
    "push_inodeblocks_del"
    (Nn_db.inodeblocks_del_e ~inode ~blkidx ~len)
    sess
    emit

let proc_push_names_ins sess (name,inode) emit =
  push
    "push_names_ins"
    (Nn_db.names_ins_e ~name ~inode)
    sess
    emit

let proc_push_names_del sess name emit =
  push
    "push_names_del"
    (Nn_db.names_del_e ~name)
    sess
    emit



let onclose_nameslave srv conn_id =
  (** If the TCP connection is closed, implicitly roll the transaction back *)
  match get_trans() with 
    | None ->
	()
    | Some t ->
	Netlog.logf `Info
	  "Rolling transaction back after TCP close";
	unset_trans();
	ignore(t#rollback())


(** Binding RPC servers: *)

let setup srv (conf:Nn_config.nn_node_config) =
  (* TODO: Pfs_rpcapi_srv.Filesystem.V1.bind_async ... *)
  Pfs_rpcapi_srv.Coordination.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_find_coordinator:(fun cname ->
			      if cname = conf#nn_clustername then
				Some(Nn_elect.get_coordinator())
			      else
				None)
    ~proc_find_namenodes:(fun cname ->
			    if cname = conf#nn_clustername then
			      Array.of_list conf#nn_nodes
			    else
			      [| |])
    ~proc_is_coordinator:(fun cname -> false)
    ~proc_clustername:(fun () -> conf#nn_clustername)
    ~proc_find_inodecaches:(fun _ -> [| |])  (* no such thing on a slave so far *)
    ~proc_local_identities:(fun _ -> [| |])  (* not supported here *)
    srv;

  let no_dn_admin name _ =
    failwith ("Dn_admin." ^ name ^ ": Not available on slave node") in

  Pfs_rpcapi_srv.Dn_admin.V1.bind
    ~proc_null:(fun () -> ())
    ~proc_add_datanode:(no_dn_admin "add_datanode")
    ~proc_enable_datanode:(no_dn_admin "enable_datanode")
    ~proc_disable_datanode:(no_dn_admin "disable_datanode")
    ~proc_is_enabled:(no_dn_admin "is_enabled")
    ~proc_lookup:(no_dn_admin "lookup")
    ~proc_datanodes:(no_dn_admin "datanodes")
    ~proc_destroy_datanode:(no_dn_admin "destroy_datanode")
    srv;

  Pfs_rpcapi_srv.Nameslave.V1.bind_async
    ~proc_null:(fun _ _ emit -> emit ())
    ~proc_begin_transaction:(proc_begin_transaction conf)
    ~proc_prepare_commit
    ~proc_commit
    ~proc_push_datastore_upd
    ~proc_push_datastore_del
    ~proc_push_revision_upd
    ~proc_push_blockalloc_upd
    ~proc_push_inode_ins
    ~proc_push_inode_upd
    ~proc_push_inode_upd_time
    ~proc_push_inode_del
    ~proc_push_inodeblocks_ins
    ~proc_push_inodeblocks_del
    ~proc_push_names_ins
    ~proc_push_names_del
    srv;

  Rpc_server.set_onclose_action srv (onclose_nameslave srv)
